// @_lostbuoy_
// by Mark Sample
// https://twitter.com/_LostBuoy_

// Artist's Statement

// This bot speculates on the thoughts of a real National Oceanic and Atmospheric Administration buoy
// that has come adrift in the Pacific Ocean. The bot uses real-time weather and oceanic data transmitted
// by the buoy. This data is publicly available at http://www.ndbc.noaa.gov/station_page.php?station=51000
// The buoy is failing and occasionally errors occur in its data collection. @_lostbuoy_ attempts to 
// dramatize these data errors by glitching itself. In the final analysis, @_lostbuoy_ asks two questions:
// "What is it like to be a thing?" (a la Bogost) and
// "What is it like to be a broken thing?" (a la Steven Jackson)


// Copyright (C) 2014 by Mark Sample

// Permission is hereby granted, free of charge, to any person obtaining a copy of this software
// and associated documentation files (the "Software"), to deal in the Software without restriction,
// including without limitation the rights to use, copy, modify, merge, publish, distribute, 
// sublicense, and/or sell copies of the Software, and to permit persons to whom the Software 
// is furnished to do so, subject to the following conditions:

// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING 
// BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE 
// AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

// **** now the fun can begin ****

// Let's get the required libraries

var ent = require('ent'); // This does HTML en/decoding, changing HTML like &amp; into &
var Twit = require('twit'); // For the posting to Twitter. A RPITA
var rita = require('./rita.js'); // The RiTa library provides all the natural language processing, including Markov chains
var countdown = require('./countdown.js'); // This is how @_lostbuoy_ knows how long it's been adrift

// We also need a way to grab real-time data from the National Data Buoy Center
// There's no API, but there is an hourly RSS feed (okay, so not quite real-time)
// The next few lines grab that feed

var FeedParser = require('feedparser')
  , request = require('request');

var req = request('http://www.ndbc.noaa.gov/data/latest_obs/51000.rss')
  , feedparser = new FeedParser();

// The first part of every _lostbuoy_ tweet comes from the buoy data
// The second part of every _lostbuoy_ tweet is a deformation of Moby-Dick
// Made by Markov-chaining the entire text
  
  var fs = require('fs');

   rm = new rita.RiMarkov(3), file = 'moby-dick.txt'; // Loads the entire text of Moby-Dick into the Markov chain generator
   // The (3) above sets the Markov chain length to 3. A bigger number (like 4 or 5) would mean the sentences are more
   // likely to be coherent. A smaller number (e.g. 2) would generate more nonsensical sentences, as the engine
   // would only look for bigrams as the measure of natural language occurrence.
   
   rm.minSentenceLength = 2; // Sets the minimum sentence length to two words. Good for exclamations like "Ho, gluepots!"
   rm.maxSentenceLength = 10; // Sets the maximum sentence length to ten words. I'd like to go longer, but need to stay under then 140 character limit
  
var adrift  = countdown( new Date(2013, 2, 10),null,0,2 ); // Calculates how long the buoy has been adrift

// The next function is used occasionally to scramble (i.e. glitch) the weather data.
// Inside info: The buoy's RSS feeds are inconsistently formatted, thanks to intermittent sensor errors
// For example, about 1 out of 5 times the sensor doesn't transmit the wave height 

function scramble( s ) {
 return s.replace(
  /\b([a-z])([a-z]+)([a-z])\b/gi,
  function( t, a, b, c ) {
   b = b.split( /\B/ );
   for( var i = b.length, j, k; i; j = parseInt( Math.random() * i ),
    k = b[--i], b[i] = b[j], b[j] = k ) {}
   return a + b.join( '' ) + c;
  }
 );
}

// All the OAuth keys to send posts to Twitter 

var T = new Twit({
  consumer_key:         '', 
  consumer_secret:      '',
  access_token:         '',
  access_token_secret:  ''
});

// Now here's the main event

function makeLost() 

	{

		var loArray = ["Lo","Oh","Eh","Ye","Ah","Aye","Sir","Lad","Lass"]; // A bunch of exclamations commonly used in Moby-Dick
		var lo = loArray[Math.floor(Math.random() * loArray.length)];

		
// Next comes a bunch of code that grabs the RSS feed and parses it
// I'm essentially taking RSS and converting it into something approaching JSON
// I had originally used Yahoo Pipes to do this, but it ended up being
// just as unreliable as the buoy itself.
		
req.on('error', function (error) {
  // handle any request errors
});
req.on('response', function (res) {
  var stream = this;

  if (res.statusCode != 200) return this.emit('error', new Error('Bad status code'));

  stream.pipe(feedparser);
});


feedparser.on('error', function(error) {
  // always handle errors
});
feedparser.on('readable', function() {
  // This is where the action is!
  var stream = this
    , meta = this.meta // **NOTE** the "meta" is always available in the context of the feedparser instance
    , item;

  while (item = stream.read()) {
    var buoydata = item.description
	buoydata = buoydata.replace(/(<([^>]+)>)/ig,"");

	var buoyArray = buoydata.split("\n");
	console.log(buoyArray.length);
	if (buoyArray.length <= 9) { 
		
	// The above like determines whether the buoy is currently fully functioning or not.
	// When the buoy's sensors are functioning properly, it transmits 12 pieces of data:
	// 1. Date
	// 2. Location (a static location where the buoy is supposed to be, not its current location.
	//	  There is no GPS on-board the buoy)
	// 3. Wind direction
	// 4. Wind speed (average)
	// 5. Wind gust (peak speed)
	// 6. Significant wave height
	// 7. Dominant wave period (max time between crests of waves)
	// 8. Average wave period (average time between waves)
	// 9. Atmospheric Pressure
	// 10. Pressure Tendency (is air pressure rising or falling)
	// 11.Air temperature
	// 12. Water temperature
	//
	// If the buoy is not functioning properly, it will have less than 10 of these fields.
	// That's what "if (buoyArray.length <= 9)" figures outs. 
	// The following cases apply when the buoy is not transmitting full data

	// We now javascript's "switch" function to randomly choose one of six templates to tweet
	// Each template uses a different piece of data from the buoy
	// It avoids using the broken data, but I still wanted to acknowledge the presence
	// of bad data. So Case 5 below blatantly announcenes the presence of an error.
	// While Case 6 thematizes error by using the scramble function to make a textual glitch.
	
				switch (Math.floor(Math.random() * 7)) {

				case 0: // Date

					var calendar = buoyArray[0];	
					update = calendar + ". Adrift for " + adrift.toString().replace(",","") + ".";
						
				break;
				
				case 1: // Location
				
					var location = buoyArray[1];
						location = location.trim().replace("Location: ","").replace(" ",", ");
						update = "Last known location: " + location + ". But are coordinates a location?";
				
				break;
					
				case 2: // wind direction (plus some regex to make output more readable)
				
					var winddir = buoyArray[2];
						winddir = winddir.trim().replace("Wind Direction: ", "").replace(/\s+\(.*?\)/, "");
						
												if (winddir.length >= 3) 
							{
							winddirCardinal1 = winddir.substr(0,1);
							winddirCardinal2 = winddir.substr(1,2);
							winddirCardinal1 = winddirCardinal1.replace("N","north by ").replace("S","south by ").replace("E","east by ").replace("W","west by ");
							winddirCardinal2 = winddirCardinal2.replace(/N/g,"north").replace(/S/g,"south").replace(/E/g,"east").replace(/W/g,"west");
							winddir = winddirCardinal1 + winddirCardinal2;
							}
				
						else
							{
							winddir = winddir.replace(/N/,"north").replace(/S/,"south").replace(/E/,"east").replace(/W/,"west");
							}
							
					var windspeed = buoyArray[3];
						windspeed = windspeed.trim().replace("Wind Speed: ","");
					
					update = lo + ", wind blows from the " + winddir + ", at " + windspeed + ".";

				break;
				
				case 3: // Gusts??
					
					var gust = buoyArray[4];
						gust = gust.trim();
					
					commandArray = ["Feel the gusts at ", "Stare into the gusts of ","Face the heady winds of "];
					command = commandArray[Math.floor(Math.random() * commandArray.length)];
					update = "Nothing to report except " + ent.decode(gust) + ".";
						
				break;
				
				case 4: // Buoy Details
				
					discusArray = ["3-meter discus buoy","NOAA weather buoy","buoy with an AMPS payload"];
					discus = discusArray[Math.floor(Math.random() * discusArray.length)];
					update = "Station 510000 is lost.";

				break;

				case 5: // Was once Wave height, Now Data Failure
				
					var failureArray = ["Data Failure","No Valid Observation Data","Transmission Failure","System Failure","Parity Error and Truncation","Intermittent Buoy Data"];
						failure = failureArray[Math.floor(Math.random() * failureArray.length)];
					var code = '#'+Math.floor(Math.random()*16777215).toString(16); // Creates real-looking (but fake) hexadecimal error codes. Inspired by MS Windows.

					update = failure + " [Error " + code.toUpperCase() + "]";


				break;
				
				
				case 6: // Was once Air Temperature, now datadump

					update = ent.decode(buoyArray[7].trim()) + " " + ent.decode(buoyArray[3].trim()) + " " + ent.decode(buoyArray[5].trim());
					update = scramble(update);
					
				break;
				
				}

				fs.readFile(__dirname + '/' + file, function(e, data) {   
				if (e) throw e;     
				loadMarkov(data.toString());
					});

					// Next few lines appends Markov chained Moby Dick sentences to the buoy data
					
					function loadMarkov(text) {
						rm.loadText(text);
						var sentenceOne = rm.generateSentences(1);
						var sentenceTwo = rm.generateSentences(1);
						var whale = sentenceOne + " " + sentenceTwo;
							whale = whale.replace('him', 'me').replace('"','');
						update = update + " " + whale;

				console.log(buoyArray);
				console.log(update + " (" + update.length + ")");
	
					// Toggle Twitter  by un/commenting next four lines
					T.post('statuses/update', { status: update}, function(err, reply) {
					console.log("error: " + err);
					console.log("reply: " + reply);
					});
				
		}
		}
		
		else {
		
		// If the buoy's RSS feed contains more than 10 fields we can assume it's working properly.
		// The following 10 cases are then used instead of the cases above.
	
switch (Math.floor(Math.random() * 10)) {

	case 0: // Date

		var calendar = buoyArray[0];	
		update = calendar + ". Adrift for " + adrift.toString().replace(",","") + ".";
			
	break;
	
	case 1: // Location
	
		var location = buoyArray[1];
			location = location.trim().replace("Location: ","").replace(" ",", ");
		update = "Last known location: " + location + ". Now lost.";
	
	break;
		
	case 2: // wind direction
	
		var winddir = buoyArray[2];
			winddir = winddir.trim().replace("Wind Direction: ", "").replace(/\s+\(.*?\)/, "");
			
			if (winddir.length >= 3) 
				{
				winddirCardinal1 = winddir.substr(0,1);
				winddirCardinal2 = winddir.substr(1,2);
				winddirCardinal1 = winddirCardinal1.replace("N","north by ").replace("S","south by ").replace("E","east by ").replace("W","west by ");
				winddirCardinal2 = winddirCardinal2.replace(/N/g,"north").replace(/S/g,"south").replace(/E/g,"east").replace(/W/g,"west");
				winddir = winddirCardinal1 + winddirCardinal2;
				}
	
			else
				{
				winddir = winddir.replace(/N/,"north").replace(/S/,"south").replace(/E/,"east").replace(/W/,"west");
				}
				
		var windspeed = buoyArray[3];
			windspeed = windspeed.trim().replace("Wind Speed: ","");
		
		update = lo + ", wind blows from the " + winddir + ", at " + windspeed + ".";

	break;
	
	case 3: // Gusts
		
		var gust = buoyArray[4];
			gust = gust.trim().replace("Wind Gust: ","");
		
		commandArray = ["Feel the gusts at ", "Stare into the gusts of ","Face the heady winds of ","Blown off course by gusts of "];
		command = commandArray[Math.floor(Math.random() * commandArray.length)];
		update = command + gust + ".";
			
	break;
	
	case 4: // Buoy Details
	
		discusArray = ["a 3-meter discus buoy","a NOAA weather buoy","a buoy with an AMPS payload","a lost buoy","a forgotten experiment","a floating lab","the Internet of Things"];
		discus = discusArray[Math.floor(Math.random() * discusArray.length)];
		update = "This is data from " + discus + ", once moored 245 miles northeast of Hawaii.";

	break;

	case 5: // Wave height (plus some conditions to test whether we can count the waves as steep or small)
	
		var calendar = buoyArray[0];	
		var waveheight = buoyArray[5];
			waveheight = waveheight.trim().replace("Significant Wave Height: ","").replace(" ft","");
		
					if (waveheight > 10) {
					var steepness = "steep";
										 }
					else if (waveheight < 3) {
					var steepness = "slight";
										}
										
					else {
					var steepness = "average";
						}
			
				var colon = calendar.indexOf(":");
				var precolon = colon - 2;
				var hour = calendar.substring(colon,precolon).trim(); 
				
				if (calendar.indexOf("pm") != -1 && hour > 5 && hour < 12) {
				var daytime = "this evening";
				}
				else if (calendar.indexOf("am") != -1 && hour <= 4) {
				var daytime = "tonight";
				}
				else if (calendar.indexOf("am") != -1 && hour >= 5) {
				var daytime = "this morning";
				}
				else 
					{ 
					daytime = "today";
					}
		var sayArray = ["you might say","you could say","the sensors say","the data's flawed but","any sailor would know"];
		var say = sayArray[Math.floor(Math.random() * sayArray.length)];
					
		
		update = lo + ", " + say + " the waves are " + steepness + " " + daytime + ", at " + waveheight + " ft.";

	break;
	
	case 6: // Wave period
	
		var waveperiod = buoyArray[7];
			waveperiod = waveperiod.trim().replace("Average Period: ","").replace("sec","seconds");
			countArray = [", count it: ",", measure it: ",", watch it: ",", time it: "];
			count = countArray[Math.floor(Math.random() * countArray.length)];
			update = lo + count + waveperiod + " between every wave.";

	break;
	
	case 7: // Air Temperature

		var airtemp = buoyArray[10];
			airtemp = airtemp.trim().replace("Air Temperature: ","").replace(/\s+\(.*?\)/, "");
		update = "The air, " + lo.toLowerCase() + ", you breathe it and it is " + ent.decode(airtemp) + ".";
	break;

	case 8: // Water Temperature
		
		var watertemp = buoyArray[11];
			watertemp = watertemp.trim().replace("Water Temperature: ","").replace(/\s+\(.*?\)/, "");
		commandArray = ["Lower your hand into the ", "Dip your oar into the ","Aye, today there's "];
			command = commandArray[Math.floor(Math.random() * commandArray.length)];
		
		update = command + ent.decode(watertemp) + " water.";

	break;
	
	case 9: // Air Pressure, plus what I think is a clever way to detect whether it's rising or falling 
	
		var airpressure = buoyArray[8];
			airpressure = airpressure.trim().replace("Atmospheric Pressure: ","").replace(/\s+\(.*?\)/, "").replace("in","inches");
		var pressuretendency = buoyArray[9]
			pressuretendency = pressuretendency.trim().replace("Pressure Tendency: ","");
				if (pressuretendency.indexOf(0) == "+")
					{
					var tendency = "rising";
					}
				else
					{
					var tendency = "falling";
					}
			var noteArray = ["Note ","Observe ","Sensors show ","Why is "];
				note = noteArray[Math.floor(Math.random() * noteArray.length)];
					if (note == "Why is ")
						{
						var ending = "?";
						}
					else
						{
						var ending = ".";
						}
		update = note + "the air pressure at " + airpressure + " and " + tendency + ending;
	break;

	
	}
		
	fs.readFile(__dirname + '/' + file, function(e, data) {   
    if (e) throw e;     
    loadMarkov(data.toString());
});

		function loadMarkov(text) {
			rm.loadText(text);
			var sentenceOne = rm.generateSentences(1);
			var sentenceTwo = rm.generateSentences(1);
			var whale = sentenceOne + " " + sentenceTwo;
				whale = whale.replace('him', 'me').replace('"','');
			update = update + " " + whale;

	console.log(buoyArray);
	console.log(update + " (" + update.length + ")");
	
					// Toggle Twitter  by un/commenting next four lines
					T.post('statuses/update', { status: update}, function(err, reply) {
					console.log("error: " + err);
					console.log("reply: " + reply);
					});
		
		}


		}

	}
  
});
}

makeLost();

// Lines below set up a schedule to run the makeLost function. Not necessary when the program is being
// called by a cronjob

setInterval(function() {
  try {
    makeLost();
  }
  catch (e) {
    console.log(e);
  }
},1000 * 60 * 37);